// scalar product: 0: very simple approach
#include <cassert>
#include <ctime>
#include <iostream>
using namespace std;

//    kernel function on device
__global__ void scalar(double *sg, const double *x, const double *y, int N)
{
    extern __shared__ double sdata[];
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < N)
    {
        sdata[threadIdx.x] = x[idx] * y[idx];
    }
    else
        sdata[threadIdx.x] = 0.0;

    __syncthreads();
    if (threadIdx.x == 0)
    {
        double sum = sdata[0];
        for (int i = 1; i < blockDim.x; ++i)
        {
            sum += sdata[i];
            sg[blockIdx.x] = sum;
        }
    }
}


//  main function on host
int main(void)
{
    double *x_h, *y_h, *s_h; // host data
    double *x_d, *y_d, *s_d; // device data
    double sum, tstart, tgpu;
    const int N = 100000000,
              nBytes = N * sizeof(double);
    const int LOOP = 100;

    const int blocksize = 64,
              gridsize = ceil( N / static_cast<double>(blocksize)),
              resBytes = gridsize * sizeof(double),
              sBytes   = blocksize * sizeof(double) ;
    dim3 dimBlock(blocksize);
    dim3 dimGrid(gridsize);

    cout << endl << gridsize << " x " << blocksize << " Threads\n";

    x_h = new double [N];
    y_h = new double [N];
    s_h = new double [gridsize];
// allocate memory on device
    cudaError_t  err;               //  cuda error status
    err = cudaMalloc((void **) &x_d, nBytes);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));
    err = cudaMalloc((void **) &y_d, nBytes);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));
    err = cudaMalloc((void **) &s_d, resBytes);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));

    double ss = 0.0;
    for (int i = 0; i < N; i++)
    {
        x_h[i] = (i % 137) + 1;
        y_h[i] = 1.0 / x_h[i];
        ss += x_h[i] * y_h[i];
    }

// copy data:  host --> device
//cutilSafeCall(cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice));
    err = cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));
    err = cudaMemcpy(y_d, y_h, nBytes, cudaMemcpyHostToDevice);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));

    tstart = clock();
    for (int k = 0; k < LOOP; ++k)
    {
// call the kernel function with  dimGrid.x * dimBlock.x threads
        scalar <<< dimGrid, dimBlock, sBytes>>>(s_d, x_d, y_d, N);
// copy data:  device --> host
        err = cudaMemcpy(s_h, s_d, resBytes, cudaMemcpyDeviceToHost);
        if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));

        sum = 0.0;
        for (int i = 0; i < gridsize; i++) sum += s_h[i];
    }
    tgpu = (clock() - tstart) / CLOCKS_PER_SEC / LOOP;
    cout << "GPU: " << tgpu << " sec.; " <<
         2e-9 * N / tgpu << " GFLOPS\n";

    if (static_cast<int>(sum) != N)
    {
        cout << " !!  W R O N G  result  " << endl;
        cout << sum << "  vs.  " << N << "(" << ss << ")" << endl;
    }
    else
        cout << " correct  result  " << endl;

    delete [] x_h;
    delete [] y_h;
    delete [] s_h;
    cudaFree(x_d);
    cudaFree(y_d);
    cudaFree(s_d);
    return 0;
}
